/*
 * Predescu Maria 334CB
 * Proiect PM 2019 - MusiCube
 * 
*/

#include "header.h"

/* Retin un vector pentru fiecare layer de pe orizontala*/
int layers_oriz[4] = {LAYER1, LAYER2, LAYER3, LAYER4};

/* Retin un vector pentru fiecare pin la care sunt conectate laturile verticale
 * ale cubului
*/
int layer_vert[16] = {
	PA7, PA6, PA5, PA4,
	PA3, PA2, PA1, PA0,
	PC4, PC5, PC6, PC7,
	PD6, PC1, PC2, PC3
	};

int vx[4];

/* melodia curenta */
int current_id;
int *current_melody;
int *current_duration;
uint16_t current_size;
int16_t this_duration;
int model_layer;

unsigned int k;
int inner, outer;

/* initializari */
void initialize(void) {

	/* pentru cub */
	DDRA |= 0xff;
	DDRC |= 0xff;
	DDRD |= (1 << PD7) | (1 << PD0) | (1 << PD1) | (1 << PD4) | (1 << PD6);

	/* sting ledul de pe PD6 */
	PORTD &= ~(1 << PD6);

	/* Butonul PB2 este setat ca intrare */
	DDRB &= ~(1 << PB2);
	PORTB |= (1 << PB2);
	
	/* prima melodie va fi got */
	current_id = 0;
	current_melody = got;
	current_duration = got_duration;
	current_size = sizeof(got) / sizeof(int);
}

/* Functiile pentru jocurile de LED-uri */

/* Functie pentru a aprinde toate LED-urile din cub */
void cube_on(void)
{
	PORTD |= (1<<LAYER1);	
	PORTD |= (1<<LAYER2);	
	PORTD |= (1<<LAYER3);	
	PORTD |= (1<<LAYER4);	
	
	PORTD |= (1 << PD6);
	PORTA = FULL_LAYER_A;
	PORTC = FULL_LAYER_C;
}

/* Functie pentru a stinge toate LED-urile din cub */
void cube_off(void)
{
	PORTD &= ~((1<<LAYER1));	
	PORTD &= ~((1<<LAYER2));	
	PORTD &= ~((1<<LAYER3));	
	PORTD &= ~((1<<LAYER4));	
	
	PORTD &= ~(1 << PD6);
	PORTA = EMPTY_LAYER_A;
	PORTC = EMPTY_LAYER_C;
}

/* Functie pentru a aprinde un layer intreg */
void one_layer_on(int i) {
	PORTD |= (1 << layers_oriz[i]);
	
	PORTD |= (1 << PD6);
	PORTA = FULL_LAYER_A;
	PORTC = FULL_LAYER_C;
}

/* Functie pentru a stinge un layer */
void one_layer_off(int i) {
	PORTD &= ~(1 << layers_oriz[i]);
	
	PORTD &= ~(1 << PD6);
	PORTA = EMPTY_LAYER_A;
	PORTC = EMPTY_LAYER_C;
}

/* Functie care aprinde o verticala */
void one_vertical_on(int i) {
	if (i >= 4)
		return;
	
	PORTD |= (1 << layers_oriz[0]) | (1 << layers_oriz[1]) | (1 << layers_oriz[2]) |
			 (1 << layers_oriz[3]);
	
	PORTA |= (1 << layer_vert[i]) | (1 << layer_vert[i + 4]);
	if (i != 0) {
		PORTC |= (1 << layer_vert[i + 8]) | (1 << layer_vert[i + 12]);
	} else {
		PORTC |= (1 << layer_vert[i + 8]);
		PORTD |= (1 << PD6);
	}
}

/* Functie care stinge o verticala */
void one_vertical_off(int i) {
	if (i >= 4) 
		return;
		
	PORTD &= ~(1 << layers_oriz[0]); 
	PORTD &= ~(1 << layers_oriz[1]); 
	PORTD &= ~(1 << layers_oriz[2]); 
	PORTD &= ~(1 << layers_oriz[3]);
	
	PORTA &= ~(1 << layer_vert[i]); 
	PORTA &= ~(1 << layer_vert[i + 4]);

	if (i != 0) {
		PORTC &= ~(1 << layer_vert[i + 8]); 
		PORTC &= ~(1 << layer_vert[i + 12]);
	} else {
		PORTC &= ~(1 << layer_vert[i + 8]); 
		PORTD &= ~(1 << PD6);
	}
	
}

/* Functie pentru a realiza efect de ploaie */
void rain(int iterations) {
	int i, k = 3, n, iter, x;
	
	for (iter = 0; iter < iterations; iter++) {
		/* Se alege un n random */
		n = (int)((double)rand() / ((double)RAND_MAX + 1) * 4);
		for (i = 0; i < n; i++) {
			vx[i] = 0;
		}
		/* La inceput suntem pe nivelul 1 */
		PORTD |= (1 << layers_oriz[3]);
		for (i = 0; i < n; i++) {
			/* LED random */
			x = (int)((double)rand() / ((double)RAND_MAX + 1) * 16);
			vx[i] = x;
			if (x < 8) {
				PORTA |= (1 << layer_vert[x]);
			} else {
				if (x == 12) {
					PORTD |= (1 << PD6);
				} else {
					PORTC |= (1 << layer_vert[x]);
				}
			}
		}
		_delay_ms(100);
		k = 2;
		while (k >= 0) {
			/* trecem la nivelul inferior */
			PORTD |= (1 << layers_oriz[k]);
			/* intai inchidem toate LED-urile */
			PORTA = EMPTY_LAYER_A;
			PORTC = EMPTY_LAYER_C;
			PORTD &= ~(1 << PD6);
			/* aprindem doar LED-urile aprinse pe nivelul cel mai de sus */
			for (i = 0; i < n; i++) {
				if (vx[i] != 0) {
					if (vx[i] < 8) {
						PORTA |= (1 << layer_vert[vx[i]]);
					} else {
						if (vx[i] == 12) {
							PORTD |= (1 << PD6);
						} else {
							PORTC |= (1 << layer_vert[vx[i]]);
						}
					}
				}
			}
			_delay_ms(100);
			k--;
		}
		/* dupa terminarea unei iteratii se stinge intreg cubul */
		cube_off();
		_delay_ms(100);
	}
}

/* Functie care realizeaza efectul de avioane in miscare */
void moving_planes(int dir) {
	int i;
	if (dir == 1) {
		/* de jos in sus */
		for (i = 3; i >= 0; i--) {
			/* aprindem layerul respectiv */
			one_layer_on(i);
			_delay_ms(300);
			one_layer_off(i);
		}
		for (i = 0; i < 4; i++) {
			one_layer_on(i);
			_delay_ms(300);
			one_layer_off(i);
		}		
	} else {
		/* din fata in spate */
		for (i = 3; i >= 0; i--) {
			one_vertical_on(i);
			_delay_ms(300);
			one_vertical_off(i);
		}
		for (i = 0; i < 4; i++) {
			one_vertical_on(i);
			_delay_ms(300);
			one_vertical_off(i);
		}
	}
}

/* Functie care aprinde colturile cubului */
void corner_on(void) {

	/* strat jos */
	PORTD |= (1 << LAYER1);

	/* strat sus */
	PORTD |= (1 << LAYER4);
	
	/* LED-urile */
	PORTC |= (1 << PC3);
	PORTD |= (1 << PD6);
	PORTA |= (1 << PA7) | ( 1 << PA4);
	_delay_ms(300);
}

/* Functie care stinge colturile cubului */
void corner_off(void) {
	PORTD &= ~(1 << LAYER1);
	PORTD &= ~(1 << LAYER4);
	
	PORTD &= ~(1 << PD6);
	PORTA = EMPTY_LAYER_A;
	PORTC = EMPTY_LAYER_C;
	_delay_ms(300);
}

/* Functie care aprinde cubul exterior */
void outer_cube_on(void) {
	one_vertical_on(0);
	one_vertical_on(3);
	PORTA |= (1 << PA6) | (1 << PA5);
	PORTC |= (1 << PC1) | (1 << PC2);
}

/* Functie care stinge cubul exterior */
void outer_cube_off(void) {
	one_vertical_off(0);
	one_vertical_off(3);
	PORTA &= ~((1 << PA6) | (1 << PA5));
	PORTC &= ~((1 << PC1) | (1 << PC2));
}

/* Functie care aprinde cubul interior */
void inner_cube_on(void) {
	PORTD |= (1 << PD1) | (1 << PD4);
	PORTA |= (1 << PA2) | (1 << PA1);
	PORTC |= (1 << PC5) | (1 << PC6);
}

/* Functie care stinge cubul interior */
void inner_cube_off(void) {
	PORTD &= ~((1 << PD1) | (1 << PD4));
	PORTA &= ~((1 << PA2) | (1 << PA1));
	PORTC &= ~((1 << PC5) | (1 << PC6));
}

/* Initializare pentru audio */

void timer1_init() {

    /* Fast PWM cu prag pe ICR1 si prescaler 1024 */
    TCCR1B |= (1 << CS10) | (1 << WGM13) | (1 << WGM12);
    TCCR1A |= (1 << WGM11);

	TCCR1A |= (1 << COM1B1);
	TCCR1A |= (1 << COM1A1);
	TCCR1A &= ~(1 << COM1B0);

	/* output pe PD5 */
	DDRD |= (1 << PD5);

    /* modul curent numara pana la ICR1 */
    ICR1 = 16000000 / current_melody[0];

    /* duty cycle-ul este setat initial la 25% */
    OCR1B = ICR1 / 4;

    /* duty cycle-ul pentru speaker (OC1A) este la 50% */
    OCR1A = ICR1 / 2;
}

/* Functie pentru alegerea unei melodii */
void choose_melody(void) {
	current_id = (current_id + 1) % 3;
	
	switch(current_id) {
		case GAMEOFTHRONES:
			current_melody = got;
			current_duration = got_duration;
			current_size = sizeof(got) / sizeof(int);
			break;
		case MAINTHEME:
			current_melody = maintheme;
			current_duration = maintheme_duration;
			current_size = sizeof(maintheme) / sizeof(int);
			break;
		case BEVERLYHILLS:
			current_melody = line1;
			current_duration = line1_duration;
			current_size = sizeof(line1) / sizeof(int);
			break;
	}
}

/* Functie care realizeaza aprinderea LED-urilor in ritmul muzicii */
void animation_on_music(int model_number) {	
	
	inner = FALSE;
	outer = FALSE;
	
	if(OCR1A < 9000) {
		if (model_number == 0) {
			one_layer_on(0);
		} else if (model_number == 1) {
			one_vertical_on(0);
		} else {
			inner_cube_on();
			inner = TRUE;
		}
		model_layer = LAYER1;
	} else if(OCR1A < 300) {	
		if (model_number == 0) {
			one_layer_on(1);
		} else if (model_number == 1) {
			one_vertical_on(1);
		} else {
			outer_cube_on();
			outer = TRUE;
		}
		model_layer = LAYER2;
	}
	else if(OCR1A < 15000) {
		if (model_number == 0) {
			one_layer_on(2);
		} else if (model_number == 1) {
			one_vertical_on(2);
		} else {
			inner_cube_on();
			inner = TRUE;
		}
		model_layer = LAYER3;
	}
	else if (OCR1A < 20000){
		if (model_number == 0) {
			one_layer_on(3);
		} else if (model_number == 1) {
			one_vertical_on(3);
		} else {
			outer_cube_on();
			outer = TRUE;
		}
		model_layer = LAYER4;
	} else {
		cube_on();
		model_layer = 8;
	}
}

/* Functie care sterge animatia precedenta */
void animation_on_music_off(int model_number) {

	if (model_layer == 8) {
		cube_off();
	} else if (model_number == 0) {
		one_layer_off(model_layer);
	} else if (model_number == 1) {
		one_vertical_off(model_layer);
	} else {
		if (outer == TRUE) {
			outer_cube_off();
		} else {
			inner_cube_off();
		}
	}
}


/* Functia principala */
int main (void)
{
	initialize();	

	int iter_number = 0, init = 0, model_number = 0;
	char music = FALSE;
	
	k = 1;
	
	while (1) {
		if (music != TRUE) {
			/* intai animatiile */
			moving_planes(0);
			if ((PINB & (1 << PB2)) == 0) {
				cube_off();
				/* la apasarea butonului, aleg muzica */
				music = TRUE;
			}
			
			cube_on();
			_delay_ms(800);
			cube_off();
			if ((PINB & (1 << PB2)) == 0) {
				cube_off();
				/* la apasarea butonului, aleg muzica */
				music = TRUE;
			}
			
			rain(10);
			if ((PINB & (1 << PB2)) == 0) {
				cube_off();
				/* la apasarea butonului, aleg muzica */
				music = TRUE;
			}
			
			moving_planes(1);
			if ((PINB & (1 << PB2)) == 0) {
				cube_off();
				/* la apasarea butonului, aleg muzica */
				music = TRUE;
			}
			
			corner_on();
			corner_off();
			
			if ((PINB & (1 << PB2)) == 0) {
				cube_off();
				/* la apasarea butonului, aleg muzica */
				music = TRUE;
			}
		} else {
			if (init == 0) {
				
				/* pornesc redarea audio */
				timer1_init();
				sei();

				init = 1;
			}
			
			/* in functie de ICR1 canta notele */
			ICR1 = F_CPU / current_melody[k];
			OCR1A = ICR1 / 2;
			OCR1B = ICR1 / 4;
			
			/* alegem modelul */
			animation_on_music(model_number);

			_delay_ms(200);
			k++;
			if (k == current_size) {
				k = 0;
			}
			
			/* sting ce am aprins inainte */
			animation_on_music_off(model_number);
			
			/* la apasarea pe buton se schimba melodiile */
			if ((PINB & (1 << PB2)) == 0) {
				choose_melody();
			}
			
			iter_number++;
			/* la 50 de iteratii se aprind altfel LED-urile */
			if (iter_number == 50) {
				model_number = (model_number + 1) % 3;
				iter_number = 0;
			}
		}
	}

	return 0;
}